/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WOBJECT_H
#define WOBJECT_H

#include "fwk.h"
#include <iostream>

namespace fwk
{

/**
 * @brief Base class of all virtual objects
 *
 *        The only purpose of this class is that it avoids many copy/paste for
 *        the print function (__str__() fct) in SWIG/python
 */

class FWK_API wObject
{
public:
    wObject() = default;
    virtual ~wObject() {}
    wObject(const wObject &) = delete;
    wObject &operator=(const wObject &) = delete;

#ifndef SWIG
    friend FWK_API std::ostream &operator<<(std::ostream &out, wObject const &obj);
    virtual void write(std::ostream &out) const;
#endif
};

/**
 * @brief Base class of smart-pointed objects (shared_ptr)
 *
 *        This base class is only required for "print" in SWIG/python.
 *        std::cout << *o works from C++ (it calls "write") even if the object inherits from wObject
 *        which is not defined in SWIG as shared_ptr-managed object
 *        However "print o" fails in python if the shared object inherits from wObject.
 *        This class also avoids a warning in SWIG telling that some derived or base classes are not
 *        managed by shared_ptr.
 */

class FWK_API wSharedObject
{
public:
    wSharedObject() = default;
    virtual ~wSharedObject() {}
    wSharedObject(const wSharedObject &) = delete;
    wSharedObject &operator=(const wSharedObject &) = delete;

#ifndef SWIG
    friend FWK_API std::ostream &operator<<(std::ostream &out, wSharedObject const &obj);
    virtual void write(std::ostream &out) const;
#endif
};

} // namespace fwk

#endif // WOBJECT_H
