/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WFCT2_H
#define WFCT2_H

#include "tbox.h"
#include "wObject.h"
#include <memory>
#include <vector>
#include <Eigen/Dense>

namespace tbox
{

/**
 * @brief Matrix functions to be integrated over an element
 *
 * eval() evaluates the function at kth Gauss point
 * @authors Romain Boman
 */
class TBOX_API Fct2 : public fwk::wSharedObject
{
public:
#ifndef SWIG
    Fct2() : wSharedObject()
    {
    }
#endif
    virtual ~Fct2()
    {
    }

    virtual void evalall() {}
#ifndef SWIG
    virtual void eval(Element const &e,
                      std::vector<double> const &u,
                      size_t k, Eigen::MatrixXd &out, bool fake) const = 0;
#endif
};

/**
 * @brief matrix function M = Cte
 */
class TBOX_API Fct2C : public Fct2
{
    Eigen::Matrix2d v;

public:
    Fct2C(double _v11, double _v22, double _v12);

#ifndef SWIG
    virtual void eval(Element const &e,
                      std::vector<double> const &u,
                      size_t k, Eigen::MatrixXd &out, bool fake) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief matrix function M = fct(x,y,z)
 */
class TBOX_API Fct2XYZ : public Fct2
{
public:
    virtual void eval(Eigen::Vector3d const &pos, Eigen::MatrixXd &v, bool fake) const = 0;

#ifndef SWIG
    virtual void eval(Element const &e,
                      std::vector<double> const &u,
                      size_t k, Eigen::MatrixXd &out, bool fake) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief matrix function M = fct(u)
 */
class TBOX_API Fct2U : public Fct2
{
public:
    virtual void eval(double u, Eigen::MatrixXd &v, bool fake) const = 0;

#ifndef SWIG
    virtual void eval(Element const &e,
                      std::vector<double> const &u,
                      size_t k, Eigen::MatrixXd &out, bool fake) const override;
    virtual void write(std::ostream &out) const override;
#endif
};
/**
 * @brief matrix function M = fct(u, gradU)
 */
class TBOX_API Fct2UdU : public Fct2
{
public:
    virtual void eval(Element const &e, size_t k, double u, Eigen::VectorXd const &gradu,
                      Eigen::MatrixXd &v, bool fake) const = 0;
#ifndef SWIG
    virtual void eval(Element const &e,
                      std::vector<double> const &u,
                      size_t k, Eigen::MatrixXd &out, bool fake) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief matrix function M = fct(u)
 */
class TBOX_API Fct2PwLf : public Fct2U
{
    std::shared_ptr<PwLf> v11;
    std::shared_ptr<PwLf> v22;
    std::shared_ptr<PwLf> v12;

public:
    Fct2PwLf(std::shared_ptr<PwLf> _v11, std::shared_ptr<PwLf> _v22, std::shared_ptr<PwLf> _v12) : Fct2U(), v11(_v11), v22(_v22), v12(_v12) {}

    virtual void eval(double u, Eigen::MatrixXd &v, bool fake) const override;

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace tbox

#endif // WFCT2_H
