/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wGmshImport.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include "wHex8.h"
#include "wTetra4.h"
#include "wQuad4.h"
#include "wTri3.h"
#include "wLine2.h"
#include "wPoint1.h"
#include "wResults.h"

using namespace tbox;

GmshImport::GmshImport(std::shared_ptr<MshData> _msh) : MshImport(_msh)
{
}

GmshImport::~GmshImport()
{
    std::cout << "~GmshImport()\n";
    nodemap.clear();
}

/**
 * @brief Load a mesh from a file
 */
void GmshImport::load(std::string const &fname, int myrank)
{
    // clear data structure
    msh->clear();
    nodemap.clear();

    std::cout << "reading file " << fname << "...\n";

    // read filename
    auto idx = fname.find(".msh");
    if (idx != std::string::npos)
    {
        msh->name = fname.substr(0, idx);
        std::cout << "\tsetting mesh name to \"" << msh->name << "\"\n";
    }

    // read gmsh .msh file with stdio commands
    FILE *file = fopen(fname.c_str(), "r");
    if (!file)
        throw std::runtime_error("Error opening file (load)");

    int skipped = 0;
    while (1)
    {
        // read next line
        char line[128];
        if (fgets(line, sizeof line, file) == NULL)
            break;

        // process code
        if (searchCode("$MeshFormat", line) == 0)
            readMshFormat(file);
        else if (searchCode("$Nodes", line) == 0)
            readMshNodes(file);
        else if (searchCode("$Elements", line) == 0)
            readMshElements(file, myrank);
        else if (searchCode("$PhysicalNames", line) == 0)
            readMshPhyNames(file);
        else
        {
            skipped += 1;
            if (verb)
                std::cout << "\t"
                          << "[code] skipping line:\"" << line << "\"\n";
        }
    }
    if (verb)
        std::cout << skipped << " lines skipped\n";

    // close file
    fclose(file);
    nodemap.clear();
    std::cout << "done" << std::endl;
}

/**
 * @brief Read results from a file at time step index n
 */
void GmshImport::load(std::string const &fname, Results &r, int n)
{
    // Sanity checks
    if (msh == nullptr)
        throw std::runtime_error("tbox::GmshImport::load mesh data structure is empty. Please load a mesh before reading the results!\n");

    // Create index maps
    std::map<int, size_t> nimap, eimap;
    for (size_t i = 0; i < msh->nodes.size(); ++i)
        nimap[msh->nodes[i]->no] = i;
    for (size_t i = 0; i < msh->elems.size(); ++i)
        eimap[msh->elems[i]->no] = i;

    // Open file
    std::cout << "reading file " << fname << "...\n";
    FILE *file = fopen(fname.c_str(), "r");
    if (!file)
        throw std::runtime_error("Error opening file (load)");

    // Read results
    int skipped = 0;
    while (1)
    {
        // read next line
        char line[128];
        if (fgets(line, sizeof line, file) == NULL)
            break;

        // process code
        if (searchCode("$MeshFormat", line) == 0)
            readMshFormat(file);
        else if (searchCode("$NodeData", line) == 0)
            readResNodes(file, line, nimap, r, n);
        else if (searchCode("$ElementData", line) == 0)
            readResElements(file, line, eimap, r, n);
        else
        {
            skipped += 1;
            if (verb)
                std::cout << "\t"
                          << "[code] skipping line:\"" << line << "\"\n";
        }
    }
    if (verb)
        std::cout << skipped << " lines skipped\n";

    // close file
    fclose(file);
    std::cout << "done" << std::endl;
}

/**
 * @brief Search for a line in a code
 */
int GmshImport::searchCode(char const *code, char const *line) const
{
    return strncmp(code, line, strlen(code));
}

/**
 * @brief Read mesh format type (version)
 */
void GmshImport::readMshFormat(FILE *file)
{
    double version = 0.0;
    int ftype = 0, dsize = 0;
    if (fscanf(file, "%lf %d %d\n", &version, &ftype, &dsize) != 3)
        throw std::runtime_error("Error reading file (readMshFormat)");
    if (ftype != 0)
        throw std::runtime_error("tbox::GmshImport::readMshFormat cannot read binary format!\n");
    double eps = 1e-5;
    if (version < 2.2 - eps || version > 2.2 + eps)
    {
        std::stringstream str;
        str << "bad .msh version: (" << version << ")";
        throw std::runtime_error(str.str());
    }
    else
        std::cout << "\treading .msh format " << version << "\n";
}

/**
 * @brief Read the nodes
 */
void GmshImport::readMshNodes(FILE *file)
{
    // read number of nodes
    int nbnod = 0;
    if (fscanf(file, "%d\n", &nbnod) != 1)
        throw std::runtime_error("Error reading number of nodes (readMshNodes)");
    std::cout << "\tcounted " << nbnod << " nodes\n";

    // allocate memory and read the nodes
    msh->nodes.reserve(nbnod);
    for (int i = 0; i < nbnod; ++i)
    {
        int no = 0;
        double x = 0.0, y = 0.0, z = 0.0;
        if (fscanf(file, "%d %lf %lf %lf\n", &no, &x, &y, &z) != 4)
            break;
        if (verb)
            std::cout << "\t" << no << ((i == nbnod - 1) ? ".\n" : ",");
        Node *nod = new Node(no, Eigen::Vector3d(x, y, z), i);
        msh->nodes.push_back(nod);

        nodemap[no] = nod;
    }
}

/**
 * @brief Read the elements
 */
void GmshImport::readMshElements(FILE *file, int myrank)
{
    // read number of elements
    int nbelm = 0;
    if (fscanf(file, "%d\n", &nbelm) != 1)
        throw std::runtime_error("Error reading number of elements (readMshElements)");
    std::cout << "\tcounted " << nbelm << " elements\n";

    // allocate memory and read the elements
    msh->elems.reserve(nbelm);
    for (int i = 0; i < nbelm; ++i)
    {
        // read number, type and nb of tags
        int no = 0, type = 0, nbtags = 0;
        if (fscanf(file, "%d %d %d", &no, &type, &nbtags) != 3)
            throw std::runtime_error("Error reading &no, &type, &nbtags (readMshElements)");
        if (verb)
            std::cout << "\t" << no << ((i == nbelm - 1) ? ".\n" : ",");

        // read/create/link tags
        // physical tag (group)
        Tag *ptag = nullptr;
        if (nbtags > 0)
            ptag = addTag(file, msh->ptags);
        // elementary tag (geometry)
        Tag *etag = nullptr;
        if (nbtags > 1)
            etag = addTag(file, msh->etags);
        // partitions
        bool belongsToPart = false;
        std::vector<Tag *> prts;
        if (nbtags > 2)
        {
            // read number of part
            int nbpart = 0;
            if (fscanf(file, "%d", &nbpart) != 1)
                throw std::runtime_error("Error reading &nbpart (readMshElements)");
            for (int j = 0; j < nbpart; ++j)
            {
                Tag *tag = addTag(file, msh->parts);
                prts.push_back(tag);
                if (myrank + 1 == tag->no || (myrank + 1 == -(tag->no)))
                    belongsToPart = true;
            }
        }
        if (static_cast<ElType>(type) == ElType::POINT1)
            belongsToPart = true;

        bool skipline = true;
        // read/link nodes (if serial run or elements belong to partition)
        if (myrank < 0 || belongsToPart)
        {
            skipline = false;
            std::vector<Node *> nods;
            if (static_cast<ElType>(type) == ElType::HEX8) // 8-node hexa
            {
                readElNods(file, 8, nods);
                msh->elems.push_back(new Hex8(no, ptag, etag, prts, nods));
            }
            else if (static_cast<ElType>(type) == ElType::QUAD4) // 4-node quad
            {
                readElNods(file, 4, nods);
                msh->elems.push_back(new Quad4(no, ptag, etag, prts, nods));
            }
            else if (static_cast<ElType>(type) == ElType::TETRA4) // 4-node tetra
            {
                readElNods(file, 4, nods);
                msh->elems.push_back(new Tetra4(no, ptag, etag, prts, nods));
            }
            else if (static_cast<ElType>(type) == ElType::LINE2) // 2-node line
            {
                readElNods(file, 2, nods);
                msh->elems.push_back(new Line2(no, ptag, etag, prts, nods));
            }
            else if (static_cast<ElType>(type) == ElType::POINT1) // 1-node point
            {
                readElNods(file, 1, nods);
                msh->elems.push_back(new Point1(no, ptag, etag, prts, nods));
            }
            else if (static_cast<ElType>(type) == ElType::TRI3) // 3-node triangle
            {
                readElNods(file, 3, nods);
                msh->elems.push_back(new Tri3(no, ptag, etag, prts, nods));
            }
            else
                skipline = true;
        }
        if (skipline)
        {
            // skip line
            char line[128];
            if (fgets(line, sizeof line, file) == NULL)
                throw std::runtime_error("Error skipping line (readMshElements)");
            if (verb)
                std::cout << "[elm] skipping line:\"" << line << "\"\n";
        }
    }
}

/**
 * @brief Read physical group names
 */
void GmshImport::readMshPhyNames(FILE *file)
{
    // read number of names
    int nbtag = 0;
    if (fscanf(file, "%d\n", &nbtag) != 1)
        throw std::runtime_error("Error reading number of physical names (readMshPhyNames)");

    // read the names
    for (int i = 0; i < nbtag; ++i)
    {
        // read dimension, label and name
        int dim = 0, no = 0;
        char pname[128];
        if (fscanf(file, "%d %d \"%[^\"]\"", &dim, &no, pname) != 3)
            throw std::runtime_error("Error reading &dim, &no, &name (readMshPhyNames)");
        // add tag to gtag list
        auto it = msh->ptags.find(no);
        if (it == msh->ptags.end())
            msh->ptags[no] = new Tag(no, pname, dim);
        msh->ntags[pname] = msh->ptags[no];
    }
}

/**
 * @brief Read nodes of an element
 */
void GmshImport::readElNods(FILE *file, int n, std::vector<Node *> &nods)
{
    int nn;
    for (int j = 0; j < n; ++j)
    {
        if (fscanf(file, "%d", &nn) != 1)
            throw std::runtime_error("Error reading element/node links (readElNods)");

        auto it = nodemap.find(nn);
        if (it == nodemap.end())
            throw std::runtime_error("Unknown node label (readElNods)");
        nods.push_back(it->second);
    }
}

/**
 * @brief Create a tag for each physical group
 */
Tag *GmshImport::addTag(FILE *file, std::map<int, Tag *> &tmap)
{
    int tag;
    if (fscanf(file, "%d", &tag) != 1)
        throw std::runtime_error("Error reading tags (addtag)");
    auto it = tmap.find(tag);
    if (it == tmap.end())
        tmap[tag] = new Tag(tag);
    return tmap[tag];
}

/**
 * @brief Read header of results
 */
void GmshImport::readResHeader(FILE *file, char *line, char *vname, unsigned long &nd, int &nc, int &nt, double &time)
{
    // Read the header
    int nstrtags = 0, nrealtags = 0, ninttags = 0; // number of string, real and integer tags
    // get the number of string tags
    if (fscanf(file, "%d", &nstrtags) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // get the (first) name of the view (variable), and skip the others if any
    if (fscanf(file, " \"%[^\"]\"", vname) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    for (int i = 0; i < nstrtags - 1; ++i)
        if (fgets(line, sizeof line, file) == NULL)
            throw std::runtime_error("tbox::GmshImport::readResHeader end-of-file reached before all lines could be read!\n");
    // get the number of real tags
    if (fscanf(file, "%d", &nrealtags) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    if (nrealtags != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // get the time
    if (fscanf(file, "%lg", &time) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // get the number of integer tags
    if (fscanf(file, "%d", &ninttags) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // get the time step index
    if (fscanf(file, "%d", &nt) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // get the number of components
    if (fscanf(file, "%d", &nc) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // get the number of data
    if (fscanf(file, "%lu", &nd) != 1)
        throw std::runtime_error("tbox::GmshImport::readResHeader error reading file!\n");
    // skip the number of partition, if given
    if (ninttags == 4)
        if (fgets(line, sizeof line, file) == NULL)
            throw std::runtime_error("tbox::GmshImport::readResHeader end-of-file reached before all lines could be read!\n");
}

/**
 * @brief Read results at nodes
 */
void GmshImport::readResNodes(FILE *file, char *line, std::map<int, size_t> &nimap, Results &r, int n)
{
    // Read header
    char vname[128];      // name of view (variable to be read)
    int nc = 0, nt = 0;   // number of components and time step index
    unsigned long nd = 0; // number of data points
    double time = 0.;     // current time
    this->readResHeader(file, line, vname, nd, nc, nt, time);

    // Perform checks and read
    // check the number of nodes
    if (nd != static_cast<unsigned long>(msh->nodes.size()))
        throw std::runtime_error("tbox::GmshImport::readResNodes the number of data to be read does not match the number of nodes in the mesh!\n");
    // check if the view name matches a variable, and if the time step index matches the requested time step
    if (nc == 1)
    {
        auto resm = r.scalars_at_nodes;
        // set time and read data
        if (resm.find(vname) != resm.end() && nt == n)
        {
            std::cout << "\treading node data \"" << vname << "\" at time step #" << nt << std::endl;
            r.nt = nt;
            r.time = time;
            for (unsigned long i = 0; i < nd; ++i)
            {
                int no = 0;
                double da = 0.;
                if (fscanf(file, "%d %lg", &no, &da) != nc + 1)
                    throw std::runtime_error("tbox::GmshImport::readResNodes error reading file!\n");
                resm.at(vname)->at(nimap[no]) = da;
            }
        }
        // skip
        else
        {
            std::cout << "\tskipping node data \"" << vname << "\" at time step #" << nt << std::endl;
            for (unsigned long i = 0; i < nd; ++i)
                if (fgets(line, sizeof line, file) == NULL)
                    throw std::runtime_error("tbox::GmshImport::readResNodes end-of-file reached before all data could be read!\n");
        }
    }
    else if (nc == 3)
    {
        auto resm = r.vectors_at_nodes;
        // set time and read data
        if (resm.find(vname) != resm.end() && nt == n)
        {
            std::cout << "\treading node data \"" << vname << "\" at time step #" << nt << std::endl;
            r.nt = nt;
            r.time = time;
            for (unsigned long i = 0; i < nd; ++i)
            {
                int no = 0;
                double da0 = 0., da1 = 0., da2 = 0.;
                if (fscanf(file, "%d %lg %lg %lg", &no, &da0, &da1, &da2) != nc + 1)
                    throw std::runtime_error("tbox::GmshImport::readResNodes error reading file!\n");
                resm.at(vname)->at(nimap[no]) = Eigen::Vector3d(da0, da1, da2);
            }
        }
        // skip
        else
        {
            std::cout << "\tskipping node data \"" << vname << "\" at time step #" << nt << std::endl;
            for (unsigned long i = 0; i < nd; ++i)
                if (fgets(line, sizeof line, file) == NULL)
                    throw std::runtime_error("tbox::GmshImport::readResNodes end-of-file reached before all data could be read!\n");
        }
    }
    else if (nc == 9)
    {
        auto resm = r.tensors_at_nodes;
        // set time and read data
        if (resm.find(vname) != resm.end() && nt == n)
        {
            std::cout << "\treading node data \"" << vname << "\" at time step #" << nt << std::endl;
            r.nt = nt;
            r.time = time;
            for (unsigned long i = 0; i < nd; ++i)
            {
                int no = 0;
                double da0 = 0., da1 = 0., da2 = 0., da3 = 0., da4 = 0., da5 = 0., da6 = 0., da7 = 0., da8 = 0.;
                if (fscanf(file, "%d %lg %lg %lg %lg %lg %lg %lg %lg %lg", &no, &da0, &da1, &da2, &da3, &da4, &da5, &da6, &da7, &da8) != nc + 1)
                    throw std::runtime_error("tbox::GmshImport::readResNodes error reading file!\n");
                resm.at(vname)->at(nimap[no]) << da0, da1, da2, da3, da4, da5, da6, da7, da8;
            }
        }
        // skip
        else
        {
            std::cout << "\tskipping node data \"" << vname << "\" at time step #" << nt << std::endl;
            for (unsigned long i = 0; i < nd; ++i)
                if (fgets(line, sizeof line, file) == NULL)
                    throw std::runtime_error("tbox::GmshImport::readResNodes end-of-file reached before all data could be read!\n");
        }
    }
    else
        throw std::runtime_error("tbox::GmshImport::readResNodes the number of components must be 1, 3, or 9!\n");
}

/**
 * @brief Read results at elements
 */
void GmshImport::readResElements(FILE *file, char *line, std::map<int, size_t> &eimap, Results &r, int n)
{
    // Read header
    char vname[128];      // name of view (variable to be read)
    int nc = 0, nt = 0;   // number of components and time step index
    unsigned long nd = 0; // number of data points
    double time = 0.;     // current time
    this->readResHeader(file, line, vname, nd, nc, nt, time);

    // Perform checks and read
    // check the number of elements
    if (nd != static_cast<unsigned long>(msh->elems.size()))
        throw std::runtime_error("tbox::GmshImport::readResElements the number of data to be read does not match the number of elements in the mesh!\n");
    // check if the view name matches a variable, and if the time step index matches the requested time step
    if (nc == 1)
    {
        auto resm = r.scalars_at_elems;
        // set time and read data
        if (resm.find(vname) != resm.end() && nt == n)
        {
            std::cout << "\treading element data \"" << vname << "\" at time step #" << nt << std::endl;
            r.nt = nt;
            r.time = time;
            for (unsigned long i = 0; i < nd; ++i)
            {
                int no = 0;
                double da = 0.;
                if (fscanf(file, "%d %lg", &no, &da) != nc + 1)
                    throw std::runtime_error("tbox::GmshImport::readResElements error reading file!\n");
                resm.at(vname)->at(eimap[no]) = da;
            }
        }
        // skip
        else
        {
            std::cout << "\tskipping element data \"" << vname << "\" at time step #" << nt << std::endl;
            for (unsigned long i = 0; i < nd; ++i)
                if (fgets(line, sizeof line, file) == NULL)
                    throw std::runtime_error("tbox::GmshImport::readResNodes end-of-file reached before all data could be read!\n");
        }
    }
    else if (nc == 3)
    {
        auto resm = r.vectors_at_elems;
        // set time and read data
        if (resm.find(vname) != resm.end() && nt == n)
        {
            std::cout << "\treading element data \"" << vname << "\" at time step #" << nt << std::endl;
            r.nt = nt;
            r.time = time;
            for (unsigned long i = 0; i < nd; ++i)
            {
                int no = 0;
                double da0 = 0., da1 = 0., da2 = 0.;
                if (fscanf(file, "%d %lg %lg %lg", &no, &da0, &da1, &da2) != nc + 1)
                    throw std::runtime_error("tbox::GmshImport::readResElements error reading file!\n");
                resm.at(vname)->at(eimap[no]) = Eigen::Vector3d(da0, da1, da2);
            }
        }
        // skip
        else
        {
            std::cout << "\tskipping element data \"" << vname << "\" at time step #" << nt << std::endl;
            for (unsigned long i = 0; i < nd; ++i)
                if (fgets(line, sizeof line, file) == NULL)
                    throw std::runtime_error("tbox::GmshImport::readResNodes end-of-file reached before all data could be read!\n");
        }
    }
    else if (nc == 9)
    {
        auto resm = r.tensors_at_elems;
        // set time and read data
        if (resm.find(vname) != resm.end() && nt == n)
        {
            std::cout << "\treading element data \"" << vname << "\" at time step #" << nt << std::endl;
            r.nt = nt;
            r.time = time;
            for (unsigned long i = 0; i < nd; ++i)
            {
                int no = 0;
                double da0 = 0., da1 = 0., da2 = 0., da3 = 0., da4 = 0., da5 = 0., da6 = 0., da7 = 0., da8 = 0.;
                if (fscanf(file, "%d %lg %lg %lg %lg %lg %lg %lg %lg %lg", &no, &da0, &da1, &da2, &da3, &da4, &da5, &da6, &da7, &da8) != nc + 1)
                    throw std::runtime_error("tbox::GmshImport::readResElements error reading file!\n");
                resm.at(vname)->at(eimap[no]) << da0, da1, da2, da3, da4, da5, da6, da7, da8;
            }
        }
        // skip
        else
        {
            std::cout << "\tskipping element data \"" << vname << "\" at time step #" << nt << std::endl;
            for (unsigned long i = 0; i < nd; ++i)
                if (fgets(line, sizeof line, file) == NULL)
                    throw std::runtime_error("tbox::GmshImport::readResNodes end-of-file reached before all data could be read!\n");
        }
    }
    else
        throw std::runtime_error("tbox::GmshImport::readResElements the number of components must be 1, 3, or 9!\n");
}

void GmshImport::write(std::ostream &out) const
{
    out << "GmshImport on " << *msh << std::endl;
}
