/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wLine2.h"
#include "wSfLine2.h"
#include "wGaussLine2.h"
#include "wCacheLine2.h"
#include "wNode.h"

using namespace tbox;

Line2::Line2(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Line2 elements
 */
Cache &Line2::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Initialize precomputed values
 */
void Line2::initValues(bool isvol, int ordr)
{
    // Set flags
    isVol = isvol;
    order = ordr;
    hasVals = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    Js.resize(ngp);
    iJs.resize(ngp);
    detJs.resize(ngp);
    this->update();
}
/**
 * @brief Initialize precomputed gradients
 */
void Line2::initGradients()
{
    // Set flags
    hasGrads = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    gradJs.resize(ngp);
    gradDetJs.resize(ngp);
    this->update();
}
/**
 * @brief Update precomputed values and gradients
 */
void Line2::update()
{
    // Update values
    if (hasVals)
    {
        // Jacobian
        if (isVol)
            throw std::runtime_error("Line2::update(): the problem must always have a dimension less than Line2!\n");
        // Line2 is never of same dimension as Problem (surface element)
        this->buildSurfaceJ();
        // Length and CG
        this->computeV();
        this->computeCg();
        // Normal
        this->computeNormal();
    }
    // Update gradients
    if (hasGrads)
    {
        // Jacobian
        if (isVol)
            throw std::runtime_error("Line2::updateGradient(): the problem must always have a dimension less than Line2!\n");
        // Line2 is never of same dimension as Problem (surface element)
        this->buildGradientSurfaceJ();
        // Length
        this->buildGradientV(2);
        // Normal
        this->computeGradientNormal();
    }
}

/**
 * @brief Compute Jacobian determinant
 *
 * detJ = norm(x_i * dN_i)
 */
void Line2::buildSurfaceJ()
{
    for (size_t k = 0; k < getCache(order).getVGauss().getN(); ++k)
        detJs[k] = this->buildTangents(k)[0].norm();
}

/**
 * @brief Compute Jacobian determinant gradient with respect to each node coordinate
 *
 * dDetJ[l = i * 2 + j] = detJ_l * dt0_l;
 */
void Line2::buildGradientSurfaceJ()
{
    CacheLine2 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        // Shape functions, tangent vectors and determinant
        Eigen::MatrixXd const &dsf = getCache(order).getDsf(k);
        Eigen::Vector3d detJ = this->buildTangents(k)[0].normalized();

        // Jacobian gradient
        gradDetJs[k].resize(2 * 2); // 2 nodes and 2 dimensions (x, y)
        for (size_t i = 0; i < 2; ++i)
        {
            for (size_t j = 0; j < 2; ++j)
            {
                Eigen::Vector3d dt0 = Eigen::Vector3d::Zero();
                dt0(j) = dsf(0, i); // retain sf gradients associated to ith node for dimension j
                gradDetJs[k][i * 2 + j] = detJ.dot(dt0);
            }
        }
    }
}

/**
 * @brief Compute the element length
 */
void Line2::computeV()
{
    CacheLine2 &cache = getCache(order);
    GaussLine2 &gauss = cache.gauss;

    // Gauss integration
    vol = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        vol += gauss.getW(k) * this->getDetJ(k);
}
/**
 * @brief Compute length gradient with respect to each node coordinate
 *
 * dV[i] = w_k * dDetJ[i]_k
 */
void Line2::buildGradientV(int dim)
{
    GaussLine2 &gauss = getCache(order).gauss;

    // Gradient of volume
    gradVol.clear();
    gradVol.resize(2 * dim, 0.); // 2 nodes
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        double w = gauss.getW(k);                         // GP weight
        std::vector<double> dDetJ = this->getGradDetJ(k); // gradient of Jacobian determinant
        for (size_t i = 0; i < dDetJ.size(); ++i)
            gradVol[i] += w * dDetJ[i];
    }
}

/**
 * @brief Compute element tangent vectors
 */
std::vector<Eigen::Vector3d> Line2::computeTangents() const
{
    return std::vector<Eigen::Vector3d>{nodes[1]->pos - nodes[0]->pos};
}

/**
 * @brief Build element tangent vectors using shape functions at Gauss point k
 */
std::vector<Eigen::Vector3d> Line2::buildTangents(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    std::vector<Eigen::Vector3d> t(1, Eigen::Vector3d::Zero());
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t[0] += (*it)->pos * dff(0, i);
    }
    return t;
}

/**
 * @brief Compute element unit normal vector
 */
void Line2::computeNormal()
{
    Eigen::Vector3d t = this->computeTangents()[0].normalized();
    normal = Eigen::Vector3d(-t(1), t(0), 0.);
}

/**
 * @brief Compute gradient of element unit normal vector with respect to each node coordinate
 */
void Line2::computeGradientNormal()
{
    // Compute normal and norm
    Eigen::Vector3d t = this->computeTangents()[0];
    Eigen::Vector3d n = Eigen::Vector3d(-t(1), t(0), 0.);
    double nn = n.norm();

    // Compute gradients
    Eigen::Matrix3d inn = (Eigen::Matrix3d::Identity() - n * n.transpose() / (nn * nn)) / nn; // gradient of n/|n|
    gradNormal.resize(2 * 2);                                                                 // 2 nodes and 2 dimensions
    gradNormal[0] = inn * Eigen::Vector3d(0, -1, 0);
    gradNormal[1] = inn * Eigen::Vector3d(1, 0, 0);
    gradNormal[2] = inn * Eigen::Vector3d(0, 1, 0);
    gradNormal[3] = inn * Eigen::Vector3d(-1, 0, 0); // outer gradient times gradient of t (pos) wtr x
}
