/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WMSHDEFORM_H
#define WMSHDEFORM_H

#include "tbox.h"
#include "wObject.h"
#include <vector>
#include <map>
#include <memory>
#include <Eigen/Sparse>

namespace tbox
{

/**
 * @brief Handle mesh deformation
 * @authors Adrien Crovato
 */
class TBOX_API MshDeform : public fwk::wSharedObject
{
private:
    size_t mshSize;                                    ///< number of nodes in the mesh
    int nDim;                                          ///< dimension of the problem (2 or 3)
    int nthreads;                                      ///< number of threads
    int verbose;                                       ///< verbosity level
    std::vector<int> rows;                             ///< unknown local index
    Group *fldGrp;                                     ///< temporary group for field
    std::vector<Group *> fxdBndGrp;                    ///< temporary group for fixed boundaries
    std::vector<Group *> movBndGrp;                    ///< temporary group for moving boundaries
    Group *symBndGrp;                                  ///< temporary group for symmetry boundary
    std::vector<Groups *> intBndGrps;                  ///< temporary group for internal boundaries
    std::vector<Element *> fldElems;                   ///< field elements
    std::vector<Element *> movBndElems;                ///< moving boundaries elements
    std::vector<Element *> symBndElems;                ///< symmetry boundary elements
    std::vector<Element *> intBndElems;                ///< internal boundaries elements
    std::vector<Node *> fxdBndNodes;                   ///< fixed boundaries nodes
    std::vector<Node *> movBndNodes;                   ///< moving boundaries nodes
    std::vector<Node *> symBndNodes;                   ///< symmetry boundary nodes
    size_t blck;                                       ///< blocked dimension for symmetry B.C.
    std::vector<std::pair<Node *, Node *>> intBndPair; ///< internal boundaries node pairs
    std::vector<double> initPos;                       ///< initial positions of nodes before mesh deformation

    void uniqueSort(std::vector<Node *> &nodes);
    Eigen::MatrixXd buildK(tbox::Element const &e, Eigen::MatrixXd const &H);

public:
    std::shared_ptr<MshData> msh;               ///< mesh
    std::shared_ptr<tbox::LinearSolver> linsol; ///< linear (inner) solver

    MshDeform(std::shared_ptr<MshData> _msh, std::shared_ptr<tbox::LinearSolver> _linsol, int _nDim, int nthrds = 1, int vrb = 1);
    virtual ~MshDeform() { std::cout << "~MshDeform()\n"; }

    void setField(std::string const &fldName);
    void addMoving(std::string const &fxdBndName);
    void addFixed(std::string const &movBndName);
    void setSymmetry(std::string const &symBndName, size_t _blck);
    void addInternal(std::string const &intBndName, std::string const &intBndName_);
    void initialize();

    void savePos();
    void build(Eigen::SparseMatrix<double, Eigen::RowMajor> &K);
    void build(Eigen::VectorXd &f);
    void deform();
#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace tbox

#endif // WMSHDEFORM_H