#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""@package docstring
Utilities for tbox
Adrien Crovato
"""

def myrange(start, stop, step):
    """Extended range with custom step
    """
    vec = []
    i = start
    while i <= stop:
        vec.append(i)
        i += step
    return vec

def plot(x, y, cfg):
    """Plot using matplotlib
    """
    import matplotlib.pyplot as plt
    # define font
    font = {'family': 'serif', 'color': 'darkred', 'weight': 'normal', 'size': 16}
    # plot results
    plt.plot(x, y, 'b-', lw=2)
    if 'xlabel' in cfg:
        plt.xlabel(cfg['xlabel'], fontdict=font)
    if 'ylabel' in cfg:
        plt.ylabel(cfg['ylabel'], fontdict=font)
    if 'title' in cfg:
        plt.title(cfg['title'], fontdict=font)
    if 'xlim' in cfg:
        plt.xlim(cfg['xlim'])
    if 'ylim' in cfg:
        plt.ylim(cfg['ylim'])
    if 'invert' in cfg and cfg['invert']:
        plt.gca().invert_yaxis()
    # display
    plt.show()

def sort(id, data):
    """Sort data array against id (line connectivity list)
    """
    import numpy as np
    # sort id vector
    id = id[id[:,0].argsort(),:]
    # deep copy of data
    _data = np.zeros((data.shape[0], data.shape[1]))
    _data[:] = data
    # sort data against id
    _max = np.argmax(data[:,0])
    data[0,:] = _data[_max,:]
    nextId = id[np.where(id[:, 0]==_max)[0][0], 0]
    for i in range(1, id.shape[0]):
        data[i,:] = _data[id[nextId,1],:]
        nextId = id[nextId,1]

def read(filename):
    """Read from file and store in data array
    """
    import io
    import numpy as np
    # read file
    fl = io.open(filename, 'r')
    label = fl.readline().split(',')
    fl.close()
    data = np.loadtxt(filename, delimiter=',', skiprows=1)
    return data

def write(data, name, frmt, dlm, hdr, cmts):
    """Write data array to file
    """
    import numpy as np
    print('writing data file ' + name + '.')
    np.savetxt(name, data, fmt=frmt, delimiter=dlm, header=hdr, comments=cmts)
